## -*-Tcl-*- (nowrap)
 # ==========================================================================
 #  Vince's Additions - an extension package for Alpha
 # 
 #  FILE: "wwwParsing.tcl"
 #                                    created: 04/30/1997 {11:04:46 am} 
 #                                last update: 05/18/2002 {05:21:22 PM} 
 #  Author: Vince Darley
 #  E-mail: <vince@santafe.edu>
 #    mail: 317 Paseo de Peralta, Santa Fe, NM 87501, USA
 #     www: <http://www.santafe.edu/~vince/>
 #     
 #  Description:
 #  
 #  Procedures to parse HTML code for use in a rendering window.  Most of
 #  this does not rely on additional procedures found in the AlphaTcl 
 #  library, and might be useful in other applications.
 #  
 #  See the "wwwVersionHistory.tcl" file for license info, credits, etc.
 # ==========================================================================
 ##

# Make sure that the wwwMode.tcl file has been loaded.
wwwMode.tcl

proc wwwParsing.tcl {} {}

namespace eval WWW {}

# ===========================================================================
# 
#  ----  #
# 
#  Parsing  #
# 

## 
 # -------------------------------------------------------------------------
 # 
 # "WWW::parseHtml" --
 #  
 # This is the real workhorse here.  Go through the string 't', and whenever
 # an html tag is found do some fancy footwork to properly register items in
 # the various caches that will be used later to properly render the page in
 # the window.  Each item in the switch should be reducing the length of the
 # original string, or making substitutions so that other tags will be
 # rendered.  Removing as many unnecessary tags before the string is passed
 # to this proc will make it all go quicker.
 # 
 # While the original string is being parsed, the cached string can be
 # queried to determine if it ends with an empty line, etc.  'Positions' are
 # recorded according to the current length of the cached text, and can be
 # properly converted once the parsed text has been inserted into the window.
 # 
 # During the parsing, we divide the text into manageable chunks.  The idea
 # here is that if the regexp fails on a string, we add the next chunk and
 # try the regexp again -- this will happen quite a bit when the segment is
 # cutting off a tag.  Do to the magic in 'WWW::wrapText', any words that are
 # cut off midstream will be joined again.
 # 
 # What is the 'ideal' size for a string to be passed through all of these
 # 'regexp' commands?  In Alpha7, it doesn't seem to matter for speed but
 # longer strings can lead to memory problems.  In Alpha8/tk, we don't
 # seem to run into any issues unless the string is longer than 5000,
 # although 500 seems to maximize the speed of rendering.  Dividing the
 # master string reduces the parsing speed of the first frameset page of
 # the Internet Explorer help file from 26.2 to 5.0 seconds !!
 # 
 # -------------------------------------------------------------------------
 ##

proc WWW::parseHtml {t title {fillLength "72"}} {
    
    if {[string length $t] < 10} {
	# We have a problem here.
	set t "Sorry, nothing found in this file."
    }
    
    global WWWmodeVars WWW::Indentation WWW::Indent WWW::DDIndent WWW::HR \
      WWW::HtmlToStyle WWW::Pre WWW::ParsingStringLength WWW::FillColumn \
      WWW::UrlSource WWW::BaseUrl

    WWW::resetCache
    set WWW::FillColumn $fillLength

    WWW::colourWord "-*-WWW-*-" red
    WWW::cacheText \r\r
    
    # Create the string for horizontal lines.
    if {[info tclversion] < 8.0} {
	set WWW::HR "   "
	set count 10
	while {$count < [set WWW::FillColumn]} {append WWW::HR "_" ; incr count}
    } else {
	set WWW::HR "   [string repeat "_" [expr {[set WWW::FillColumn] - 10}]]"
    }

    set length1  [string length $t]
    set length2  [string length $t]
    
    # We can't quite handle rendering in background windows, and even if we
    # are in the foreground, it seems as if we really need to force the issue
    # here.  (esp for Alphatk.)
    bringToFront $title
    
    set ideal 500
    while {[string length $t] > $ideal} {
	lappend segments [string range $t 0 [expr {$ideal - 1}]]
	set t [string range $t $ideal end]
    }
    lappend segments $t ; set t ""
    set tagPattern "^(\[^<\]*(<\[<>\]\[^<\]*)*)<(\[^<>\]\[^>\]*)>\[\r\n\t \]*(.*)\$"
    while {[llength $segments]} {
        append t [lindex $segments 0]
	set segments [lrange $segments 1 end]
	while {[regexp $tagPattern $t dummy first dmy html t]} {
	    set length3 [expr (100 * ($length2 - $length1)) / $length2]
	    status::msg "Rendering -- $length3 %"
	    WWW::wrapText $first
	    regsub -all {\(} $html {\\\(} html
	    switch -regexp [string toupper $html] {
		"^A\[\r\n\t \]+(\[A-Z0-9\r\n\t =\"\]+)?HREF\[\r\n\t \]*=.*" {
		    # Find the string that will be hyperlinked.  If the closing
		    # tag doesn't exist, are we in an endless loop?
		    if {[catch {WWW::findClosingTag "</A>" t name}]} {
			set t $dummy
			break
		    }
		    # Find the href link.
		    set pat ".+(HREF\[\r\n\t \]*=\[\r\n\t \"\]*\[^\r\n\t \]+)"
		    if {[regexp -nocase $pat $html allofit href]} {
			set where  [expr 1 + [string first "=" $href]]
			set href   [string range $href $where end]
			set href   [string trim $href]
			set href   [string trim [string trim $href {\"}]]
		    } else {
			set href ""
		    }
		    # Find the target of the link.
		    set pat ".+(TARGET\[\r\n\t \]*=\[\r\n\t \"\]*\[^\r\n\t \]+)"
		    if {[regexp -nocase $pat $html allofit target]} {
			set where  [expr 1 + [string first "=" $target]]
			set target [string range $target $where end]
			set target [string trim $target]
			set target [string trim [string trim $target {\"}]]
		    } else {
			set target ""
		    }
		    # Register the link.
		    regsub -all {<[^>]*>} $name {} name
		    set name   [string trim $name]
		    set base   [set WWW::BaseUrl($title)]
		    if {[catch {url::makeAbsolute $base $href} link]} {
			set link ????://$href
		    } 
		    WWW::linkWord $name "WWW::link $link" $link $target
		}
		"^A\[\r\n\t \]+NAME\[\r\n\t \]*=.*" {
		    set anchor [string range $html [expr 1+ [string first "=" $html]] end]
		    set anchor [string trim $anchor "\""]
		    WWW::cacheItem Anchors $anchor [set WWW::ParsingStringLength]
		}
		"^APPLET" {
		    set html [lindex [join $html] 0]
		    WWW::splitText t </APPLET> pre
		    if {![regexp -nocase {code *= *([^.]*)\.class} $html dummy class]} {
			set class "applet"
		    }
		    set link "\"${class}.java\""
		    WWW::linkWord "Run java $class" "WWW::link $link" $link
		}
		"^A(REA)?(\[\r\n\t \]+(SHAPE|COORDS)\[\r\n\t \]*=\[\r\n\t \"\]*\[^\r\n\t \"\]+\"?)*\[\r\n\t \]+HREF\[\r\n\t \]*=.*" {
		    regsub -nocase -all "\[\r\n\t \]+TARGET\[\r\n\t \]*=\[\r\n\t \]*\"\[^\"\]*\"" $html {} link
		    set link [string range $link [expr 1 + [string last "=" $link]] end]
		    if {![regexp -nocase {^([^<]*)(<TD>)?</A>(.*)$} $t "" name td t]} {
			continue
		    }
		    set pad ""
		    if {[string length $name] == 0} {
			set name [file rootname [string trim $link "\""]]
			set pad "\r\r"
		    } elseif {$td == "<TD>"} {
			set pad "\r"
		    }
		    WWW::linkWord $name "WWW::link $link" $link
		    WWW::cacheText $pad
		}
		"^BR\[\r\n\t \]?" {
		    WWW::ensureBreak
		}
		"^(B|BIG|CITE|DFN|EM|I|SMALL|STRONG|U|VAR)\$" {
		    if {[regexp -nocase "^(\[^<\]*)</$html>(.*)\$" $t "" name t]} {
			global WWW::HtmlToStyle
			WWW::colourWord $name \
			  [set WWW::HtmlToStyle([string toupper $html])]
		    }
		}
		"^BLOCKQUOTE" {
		    WWW::ensureEmptyLine
		    WWW::adjustIndent 3
		}
		"^/BLOCKQUOTE" {
		    WWW::ensureEmptyLine
		    WWW::adjustIndent -3
		}
		"^DD" {
		    WWW::ensureBreak
		    if {![info exists WWW::DDIndent]} {
			set WWW::DDIndent 1
			WWW::adjustIndent 3
		    } 
		}
		"^/DD" {
		    WWW::ensureBreak
		    if {[info exists WWW::DDIndent]} {
			unset WWW::DDIndent
			WWW::adjustIndent -3
		    } 
		}
		"^DL" {
		    WWW::ensureEmptyLine
		}
		"^/DL" {
		    WWW::ensureEmptyLine
		    if {[info exists WWW::DDIndent]} {
			unset WWW::DDIndent
			WWW::adjustIndent -3
		    } 
		}
		"^DT" {
		    WWW::ensureBreak
		    if {[info exists WWW::DDIndent]} {
			unset WWW::DDIndent
			WWW::adjustIndent -3
		    } 
		}
		"^/DT" {
		    WWW::ensureBreak
		}
		"^EMBED\[\r\n\t \]+" {
		    if {[regexp -nocase {src *= *"([^"]+)"} $html dummy embed]} {
			set name "???"
			regexp {[^/:]+$} $embed name
			WWW::linkWord "Embedded '$name'." "WWW::link $embed" $embed
		    }
		}
		"^HR" {
		    WWW::ensureEmptyLine
		    WWW::cacheText [set WWW::HR]\r\r
		}			
		"^H\[0-9\]" {
		    WWW::ensureEmptyLine
		    # Preliminaries
		    set oldt $t
		    set html [lindex [join $html] 0]
		    set num  [string range $html 1 end]
		    set num  [expr {[string range $html 1 end] > 2} ? 3 : $num]
		    set color $WWWmodeVars(header${num}Color)
		    set style $WWWmodeVars(header${num}Style)
		    set indent ""
		    while {$num > 1} {append indent "  " ; incr num -1}
		    set pos0 [set WWW::ParsingStringLength]
		    # Note that headings which have links embedded within them
		    # will linked but not stylized.  Anchors will be recorded.
		    # Can't deal with both embedded at the same time.
		    if {[catch {WWW::findClosingTag "</$html>" t headerString}]} {
			set t $dummy
			break
		    }
		    if {[regexp -nocase "<A\[\r\n\t \]+HREF\[\r\n\t \]*=\[\r\n\t \"\]*(\[^>\]*)\[\r\n\t \]*>(.*)<" $headerString dummy link name]} {
			# This is a link, so we'll defer
			if {![string length $name] > 200} {
			    # Make sure that nothing funky is going on.
			    WWW::cacheItem Marks ${indent}${name} $pos0 $pos0 $pos0
			} 
			set t ${headerString}<P></P><P></P>${t}
			continue
		    } elseif {[regexp -nocase "<A\[\r\n\t \]+NAME\[\r\n\t \]*=\[\r\n\t \"\]*(\[^>\]*)\[\r\n\t \"\]*>.*</A>" $headerString dummy anchor]} {
			# Make this an anchor
			set anchor [string trim $anchor {\"}]
			WWW::cacheItem Anchors $anchor [set WWW::ParsingStringLength]
			set html   [string trim $html " \""]
		    } else {
			regexp -nocase "^(\[^<\]*)</$html>(.*)\$" $oldt dummy name t
		    }
		    regsub -all "</?\[^>\]+>" $headerString {} name
		    if {[info exists name]} {
			regsub -all "((&nbsp;)+)|(\[\r\n\t \]+)" $name { } name
			set name [string trim $name]
			if {[string length $name] > 200} {
			    # Somebody is doing something funky here ...
			    set t ${headerString}${t}
			    continue
			} 
			if {[string length $name]} {
			    WWW::colourWord $name $color $style
			    set name [win::MakeTitle [quote::Unhtml $name]]
			    while {[lcontains marks $name]} {append name " "}
			    lappend marks $name
			    WWW::cacheItem Marks ${indent}${name} $pos0 $pos0 $pos0
			    lappend markPositions [list ${indent}${name} $pos0 $pos0 $pos0]
			    unset name
			}
		    } 
		    WWW::ensureEmptyLine
		}
		"^ISINDEX" {
		    WWW::ensureEmptyLine
		    WWW::cacheText [set WWW::HR]\r\r
		    if {[info exists WWW::UrlSource($title)]} {
			set url [set WWW::UrlSource($title)]
		    } else {
			# This will probably fail later.
			set url ""
		    }
		    set name "Click here to search this index."
		    set link "WWW::indexSearch $url"
		    WWW::linkWord $name $link $link
		    WWW::cacheText \r[set WWW::HR]\r\r
		}
		"^LI" {
		    WWW::ensureBreak
		    global WWW::OLcount[set WWW::Indentation]
		    if {[info exists WWW::OLcount[set WWW::Indentation]]} {
			WWW::cacheText "[string range [set WWW::Indent] 2 end][set WWW::OLcount[set WWW::Indentation]] "
			incr WWW::OLcount[set WWW::Indentation]
		    } else {
			WWW::cacheText "[string range [set WWW::Indent] 2 end] "
		    }
		}
		"^OL" {
		    WWW::adjustIndent 3
		    global WWW::OLcount[set WWW::Indentation]
		    set WWW::OLcount[set WWW::Indentation] 1
		    WWW::ensureEmptyLine
		}
		"^/OL" {
		    global WWW::OLcount[set WWW::Indentation]
		    catch {unset WWW::OLcount[set WWW::Indentation]}
		    WWW::adjustIndent -3
		    WWW::ensureEmptyLine
		}
		"^PRE" {
		    WWW::ensureEmptyLine
		    set WWW::Pre 1
		}
		"^/PRE" {
		    set WWW::Pre 0
		    WWW::ensureBreak
		}
		"^P" {
		    WWW::ensureEmptyLine
		}
		"^/P" {
		    WWW::ensureEmptyLine
		}
		"^TR" {
		    WWW::ensureEmptyLine
		}
		"^TD" {
		    WWW::ensureSpace
		}
		"^UL" {
		    WWW::ensureEmptyLine
		    WWW::adjustIndent 3
		}
		"^/UL" {
		    WWW::ensureEmptyLine
		    WWW::adjustIndent -3
		}
		"^/.*" {
		}
		default {
		    if {[regexp -nocase "^(\[^<\]*)</$html>(.*)\$" $t dummy name t]} {
			WWW::wrapText $name
		    }
		}
	    }
# 	    set length1 [expr {[string length $segments] + [string length $t]}]
	    set length1 [expr {($ideal * [llength $segments]) + [string length $t]}]
	}
    }
    WWW::wrapText $t
}

## 
 # -------------------------------------------------------------------------
 # 
 # "WWW::findClosingTag" "WWW::splitText"  --
 # 
 # Find the closest closing tag.  We assume that 't' no longer includes the
 # opening tag, although that might not be an issue.
 # 
 # 'WWW::splitText' is a variation of this, which actually performs some
 # substition in the parsing string.
 # 
 # -------------------------------------------------------------------------
 ##

proc WWW::findClosingTag {tag _t _text} {

    upvar $_t    t
    upvar $_text text

    set tagEnd [string first [string tolower "$tag"] $t]
    set TagEnd [string first [string toupper "$tag"] $t]
    if {$tagEnd == "-1" && $TagEnd == "-1"} {
# 	alertnote "Unbalanced '$tag' tag"
	error "Unbalanced '$tag' tag"
    } elseif {$tagEnd != "-1" && $TagEnd != "-1"} {
        set end [expr {$tagEnd > $TagEnd} ? $TagEnd : $tagEnd]
    } elseif {$tagEnd == "-1"} {
        set end $TagEnd
    } else {
        set end $tagEnd
    }
    set text [string range $t 0 [expr {$end - 1}]]
    set t [string range $t [expr {$end + [string length $tag]}] end]
}

proc WWW::splitText {text at prefix} {
    upvar $prefix a
    upvar $text t
    if {[set p [string first $at [string toupper $t]]] == -1} {
	set a $t
	set t ""
    } else {
	set a [string range $t 0 [expr $p -1]]
	set t [string range $t [expr $p + [string length $at]] end]
    }
}

## 
 # -------------------------------------------------------------------------
 # 
 # "WWW::adjustIndent"  --
 # 
 # Adjust all of the indenting variables, including the HR and the fill
 # column.
 # 
 # -------------------------------------------------------------------------
 ##

if {[info tclversion] < 8.0} {
    
    proc WWW::adjustIndent {amount} {
	
	global WWW::Indent WWW::Indentation WWW::HR WWW::FillColumn
	
	if {[incr WWW::Indentation $amount] < 0} {
	    # Should never happen with properly written html code.
	    set WWW::Indentation 0
	    set WWW::Indent ""
	    set WWW::HR "   "
	    set count 10
	    while {$count < [set WWW::FillColumn]} {append WWW::HR "_" ; incr count}
	} elseif {$amount < 0} {
	    set amount [expr {abs($amount)}]
	    set WWW::Indent [string range [set WWW::Indent] $amount end]
	    set WWW::HR [string range [set WWW::HR] $amount end]
	    set count 0
	    while {$count < [expr {2 * $amount}]} {
		append WWW::HR "_"
		incr count
	    }
	    incr WWW::FillColumn $amount
	} else {
	    set count 0
	    while {$count < $amount} {
		append WWW::Indent " "
		incr count
	    }
	    set count 0
	    while {$count < $amount} {
		set WWW::HR " [set WWW::HR]"
		incr count
	    }
	    set WWW::HR [string range [set WWW::HR] 0 \
	      [expr {[string length [set WWW::HR]] - ((2 * $amount) + 1)}]]
	    incr WWW::FillColumn -$amount
	}
    }

} else {
    
    proc WWW::adjustIndent {amount} {
	
	global WWW::Indent WWW::Indentation WWW::HR WWW::FillColumn
	
	if {[incr WWW::Indentation $amount] < 0} {
	    # Should never happen with properly written html code.
	    set WWW::Indentation 0
	    set WWW::Indent ""
	    set WWW::HR "   [string repeat \"_\" [expr {[set WWW::FillColumn] - 10}]]"
	} elseif {$amount < 0} {
	    set amount [expr {abs($amount)}]
	    set WWW::Indent [string range [set WWW::Indent] $amount end]
	    set WWW::HR [string range [set WWW::HR] $amount end]
	    append WWW::HR [string repeat "_" [expr {2 * $amount}]]
	    incr WWW::FillColumn $amount
	} else {
	    append WWW::Indent [string repeat " " $amount]
	    set WWW::HR [string range [set WWW::HR] 0 \
	      [expr {[string length [set WWW::HR]] - ((2 * $amount) + 1)}]]
	    set WWW::HR [string repeat " " $amount][set WWW::HR]
	    incr WWW::FillColumn -$amount
	}
    }
}

## 
 # -------------------------------------------------------------------------
 # 
 # "WWW::ensureSpace" "WWW::ensureBreak" "WWW::ensureEmptyLine"  --
 # 
 # Ensure that the end of the parsing string contains spaces, line breaks or
 # empty lines.
 # 
 # -------------------------------------------------------------------------
 ##

proc WWW::ensureSpace {} {
    
    global WWW::Pre WWW::ParsingString
    
    if {[set WWW::Pre]} {return}
    if {![regexp "\[\r\n\t \]+$" [set WWW::ParsingString]]} {
	WWW::cacheText " "
    } 
}

proc WWW::ensureBreak {} {
    
    global WWW::Pre WWW::ParsingString
    
    if {[set WWW::Pre]} {return}
    if {![regexp "\[\r\n\]+\[\t \]*$" [set WWW::ParsingString]]} {
	WWW::cacheText "\r"
    } 
}

proc WWW::ensureEmptyLine {} {
    
    global WWW::Pre WWW::ParsingString
    
    if {[set WWW::Pre]} {return}
    if {![regexp "\[\r\n\]+\[\t \]*\[\r\n\]+\[\t \]*$" [set WWW::ParsingString]]} {
	WWW::cacheText "\r\r"
    } 
}

## 
 # -------------------------------------------------------------------------
 # 
 # "WWW::wrapText"  --
 # 
 # Given a string and some preset parameters re indentation, PRE, etc, format
 # the string so that it will be rendered appropriately respecting either the
 # WWW mode var for 'fillColumn' or the width of the current window.
 # 
 # While this might be improved in efficiency, it's really 'WWW::parseHtml'
 # that takes the most time in rendering, so that main concern here really
 # should be proper indenting and filling.
 #
 # 'ext' and 'WWW::Ext' are Vince's very clever ways of adjusting the fill of
 # a line if it contains bold, outline, or shadow fonts, which take up quite
 # a bit more room -- 'ext' indicates that the current text is lengthy, while
 # 'WWW::Ext' remembers this value in case the line isn't yet filled.
 # 
 # -------------------------------------------------------------------------
 ##

proc WWW::wrapText {text {style ""}} {
    
    global WWW::ParsingString WWW::ParsingStringLength WWW::Pre WWW::HtmlToAccents \
      WWW::Ext WWW::Indent WWW::Indentation WWW::FillColumn
    
    set posBeg [set WWW::ParsingStringLength]
    # Make sure that we have an adequate length for filling.
    if {[expr {([set WWW::FillColumn] - 20) <= [set WWW::Indent]}]} {
        # This might happen with a lot of indenting, small window parameters.
	set wwwFC 20
    } else {
	set wwwFC [set WWW::FillColumn]
    }
    # Replace some common style elements first.
    regsub -all {&\#160;|&nbsp;}   $text { }  text
    regsub -all {&lt;}             $text "<"  text
    regsub -all {&gt;}             $text ">"  text
    regsub -all {&quot;}           $text {"}  text
    regsub -all {&\#8226;}         $text {}  text
    # Replace all other style elements, except for '$amp;'.  We do these at
    # the end, in case they're being used to demonstrate HTML code --
    # otherwise, the demo string will be completely converted !!
    set loop 0
    while {[regexp "&(\[^\r\n\t ;\]+);" $text dummy accent]} {
	regsub {#0+} $accent {#} accent
	if {$accent == "amp"} {
	    regsub -all $dummy $text "<<HTMLAMPCODE>>" text 
	} elseif {[info exists WWW::HtmlToAccents($accent)]} {
	    regsub -all $dummy $text [set WWW::HtmlToAccents($accent)] text
	} else {
	    if {[regsub -all {&} $accent {\\&} accent]} {
		beep
		status::msg "There is a syntax error in &${accent}\;."
	    } else {
		set accent "\\&;${accent};"
		regsub -all $dummy $text $accent text
	    }
	    incr loop
	} 
	if {$loop > 100} {
	    break
	    status::msg "There are too many unknown characters or a problem."
	}
    }
    regsub -all {<<HTMLAMPCODE>>} $text {\&} text
    if {$loop > 0} {regsub -all {&;} $text {\&} text}
    # Remove any unnecessary leading extra spaces, carriage returns.
    if {![set WWW::Pre]} {
	regsub "^\[\r\n\t \]+\$"   $text { }  text
	if {[regexp "\[\r\n\t \]+\$" [set WWW::ParsingString]]} {
	    regsub "^\[\r\n\t \]+" $text {}   text
	} else {
	    regsub "^\[\r\n\]+"    $text {}   text
	}
    }
    # Perform some initial tests.
     if {$text == ""} {
	# Nothing here!
	return [list $posBeg $posBeg]
    } elseif {[set WWW::Pre]} {
	# This is preformatted, so simply insert.
	return [WWW::cacheText $text]
    } else {
	regsub -all "\[\t\r\n \]+" $text " " text
    }
    # Prepare the text for wrapping.
    set ext [regexp {bold|outline|shadow} $style]
    if {[regexp "\[\r\n\]\$" [set WWW::ParsingString]]} {set WWW::Ext 0}
    regsub "^.*\[\r\n\]" [set WWW::ParsingString] {} thisLine
    set x [string length $thisLine]
    while {$x < [set WWW::Indentation]} {
	WWW::cacheText " "
	incr x
    }
    set fc [expr {$wwwFC - [expr {(3 + [set WWW::Ext]) / 6}]}]
    if {$x > $fc} {
	WWW::cacheText "\r[set WWW::Indent]"
	set x 0
	set WWW::Ext 0
    }
    set fc [expr {$fc - $x + 1}]
    if {$ext} {set fc [expr {(5 + 6 * $fc) / 7}]}
    if {[set lword [string first " " $text]] == -1} {
	set lword [string length $text]
    }
    if {$fc < $lword} {
	WWW::cacheText "\r[set WWW::Indent]"
	set WWW::Ext 0
	set fc [expr $wwwFC - [set WWW::Indentation]]
	if {$ext} {
	    set fc [expr {(5 + 6 * $fc) / 7}]
	    set WWW::Ext [string length $text]
	}
    }
    # Now we cache the text, 
    while {[string length $text] > $fc} {
	set f [string last " " [string range $text 0 $fc]]
	if {$f == -1} {set f $fc}
	WWW::cacheText [string trimright [string range $text 0 $f]]
	set text [string range $text [incr f] end]
	set fc [expr [set WWW::FillColumn] - [set WWW::Indentation]]
	if {$ext} {set fc [expr {(5 + 6 * $fc) / 7}]}
	if {[string length [string trim $text]]} {
	    WWW::cacheText "\r[set WWW::Indent]"
	} 
    }
    WWW::cacheText $text
    if {$ext} {set WWW::Ext [expr {[set WWW::Ext] + [string length $text]}]}
    list $posBeg [set WWW::ParsingStringLength]
}

## 
 # -------------------------------------------------------------------------
 # 
 # "WWW::linkWord" "WWW::colourWord"  --
 # 
 # Given a string and some color/hyper arguments, cache both the text and
 # the color/hyper info that will be used once the text has actually been
 # inserted.  Since the window has no positions yet, we based the positions
 # for the colors/hypers on the current length of the parsing string and 
 # we'll do some math position once the parsing is done and the text has
 # been inserted into the window.
 # 
 # -------------------------------------------------------------------------
 ##

proc WWW::linkWord {word cmd link {target ""}} {

    if {![string length [string trim $word]]} {return}

    global WWWmodeVars WWW::Visited

    set positions [WWW::wrapText $word]
    set pos0 [lindex $positions 0]
    set pos1 [lindex $positions 1]
    set link [string trim [string trim $link] {\"}]
    if {![lcontains WWW::Visited $link]} {
	set color $WWWmodeVars(linkColor)
    } else {
	set color $WWWmodeVars(visitedLinkColor)
    }
    WWW::cacheItem Colors $pos0 $pos1 $color $cmd $link $target
}

proc WWW::colourWord {word args} {

    set positions [WWW::wrapText $word [lindex $args 0]]
    set pos0 [lindex $positions 0]
    set pos1 [lindex $positions 1]
    foreach arg $args {
	WWW::cacheItem Colors $pos0 $pos1 $arg
    }
}

#  Caching  #

## 
 # -------------------------------------------------------------------------
 # 
 # "WWW::resetCache" "WWW::cacheText" "WWW::cacheItem"  --
 # 
 # Store various bits of information that will be used once the entire html
 # file has been parsed.  None of these should be called before the new
 # window (with 'title') has been created.  Calling this at the end of the
 # rendering routine will probably ease up some memory issues.
 # 
 # -------------------------------------------------------------------------
 ##

proc WWW::resetCache {} {
    
    global WWW::ParsingString WWW::ParsingStringLength WWW::ParsingCache \
      WWW::DDIndent WWW::Indentation WWW::Indent WWW::Pre WWW::Ext
    
    set WWW::ParsingString ""
    set WWW::ParsingStringLength 0
    
    catch {unset WWW::ParsingCache}
    catch {unset WWW::DDIndent}

    set WWW::Indentation 0
    set WWW::Indent      ""
    set WWW::Pre         0
    set WWW::Ext         0
}

proc WWW::cacheText {textString} {
    
    global WWW::ParsingString WWW::ParsingStringLength
    
    set length1 [string length [set WWW::ParsingString]]
    set length2 [string length [append WWW::ParsingString $textString]]
    
    list $length1 [set WWW::ParsingStringLength $length2]
}

proc WWW::cacheItem {item args} {
    global  WWW::ParsingCache
    lappend WWW::ParsingCache($item) $args
}

# ===========================================================================
# 
# .