## -*-Tcl-*-
 # ###################################################################
 #  AlphaTcl - core Tcl engine
 # 
 #  FILE: "initialize.tcl"
 #                                    created: 25/4/02 5:28:24 
 #                                last update: 05/20/2002 {10:47:54 AM} 
 #  Author: Daniel A. Steffen
 #  E-mail: <steffen@maths.mq.edu.au>
 #    mail: Macquarie University
 #          Mathematics Department, NSW 2109, Australia
 #     www: <http://www.maths.mq.edu.au/~steffen/>
 #  
 #  Description: 
 #  
 #  initialize.tcl should be the first file sourced in the AlphaTcl
 #  startup sequence when running under Alpha8/X/Tk.
 #  It can be sourced very early as it only assumes that the
 #  following global variables have been setup
 #      HOME, PREFS, alpha::platform
 #  and that the [alert] command or the [alertnote] proc exists
 #  (for startup error reporting).
 #  
 #  This file defines the procs [alpha::PreGuiStartup] and
 #  [alpha::Startup] which need to be called later in the AlphaTcl
 #  startup process by the embedding application.
 #  
 #  History
 # 
 #  modified   by  rev reason
 #  ---------- --- --- -----------
 #  2002-04-25 DAS 1.0 original
 # ###################################################################
 ##

namespace eval alpha {}

# If we have already sourced this file, then this proc will exist
if {[llength [info procs alpha::Startup]]} { return }

# We define the 'alpha::macos' variable as follows.  This basically
# tells us that we are able to use Apple-specific technology such
# as apple-events, applescript, resources, etc.
if {$tcl_platform(platform) == "macintosh"} {
    set alpha::macos 1
} elseif {$tcl_platform(platform) == "unix" && $tcl_platform(os) == "Darwin"} {
    set alpha::macos 2
} else {
    set alpha::macos 0
}

proc alpha::checkFileExists {path {isdir 0}} {
    if {![file exists $path] || \
      ($isdir && ![file isdirectory $path]) \
      || (!$isdir && ![file isfile $path])} {
	alertnote "Fatal Alpha error:\
	  Alpha cannot find its\
	  [expr {$isdir ? {directory} : {file}}]\
	  '[file join [pwd] $path]'.\
	  You should reinstall Alpha.  Goodbye."
	quit
    }
}

if {[info tclversion] < 8.0} { return }

if {$alpha::platform == "alpha" && [llength [info commands alert]]} {
    if {[llength [info commands alertnote]] > 0} {rename alertnote __alertnote}
    ;proc alertnote {args} {
	alert -t stop -c "" -o "" [lindex $args 0] [join [lrange $args 1 end] " "]
	return
    }
    
    
    if {[llength [info commands askyesno]] > 0} {rename askyesno __askyesno}
    ;proc askyesno {args} {
	if {[lindex $args 0] == "-c"} {
	    set button [alert -t caution -k "Yes" -c "No" -o "Cancel" -C other [lindex $args 1]]
	} else {
	    set button [alert -t caution -k "Yes" -c "No" -o "" -C none [lindex $args 0]]
	}
	
	return [string tolower $button]
    }
    
    
    if {[llength [info commands buttonAlert]] > 0} {rename buttonAlert __buttonAlert}
    ;proc buttonAlert {prompt args} {
	set buttonCount [llength $args]
	
	if {$buttonCount > 3} {
	    if {[llength [info commands __buttonAlert]] > 0} {
		eval __buttonAlert [list $prompt] $args
	    }
	} else {
	    set cmd [list alert -t caution -c "" -o ""]
	    if {$buttonCount > 0} {
		set defaultText [lindex $args 0]
		lappend cmd -k $defaultText
		if {[string tolower $defaultText] == "cancel"} {
		    lappend cmd -C ok -K cancel
		} 
		if {$buttonCount > 1} {
		    set cancelText [lindex $args 1]
		    lappend cmd -c $cancelText
		    if {[string tolower $cancelText] == "cancel"} {
			lappend cmd -C cancel
		    } 
		    if {$buttonCount > 2} {
			set otherText [lindex $args 2]
			lappend cmd -o $otherText
			if {[string tolower $otherText] == "cancel"} {
			    lappend cmd -C other
			} 
		    }
		}
	    }
		    
	    lappend cmd $prompt
	    
	    eval $cmd
	}
    }
} else {
    if {[info commands alert] == "alert"} {
	rename alert __alert
    }
    ;proc alert {args} {
	set opts(-h) 0
	
	getOpts {t c k o C K O}
	
	return [buttonAlert $args $opts(-k) $opts(-c) $opts(-o)]
    }
}

proc alpha::inAlphaHierarchy {filename} {
    global HOME PREFS tcl_platform
    if {$tcl_platform(platform) == "windows"} {
	catch {set filename [file join [file attributes $filename -longname]]}
    } else {
	set filename [file join $filename]
    }
    if {[file pathtype $filename] == "relative"} {
	set filen [file join [pwd] $filename]
    } else {
	set filen $filename
    }
    # Is file in HOME or PREFS?
    if {([string first [file join $HOME] $filen] != 0) \
      && !([file type [file join $HOME]] == "link" \
      && [string first [file readlink [file join $HOME]] $filen] == 0)} {
	if {[info exists PREFS] && [file exists $PREFS]} {
	    if {([string first [file join $PREFS] $filen] != 0) \
	      && !([file type [file join $PREFS]] == "link" \
	      && [string first [file readlink [file join $PREFS]] $filen] == 0)} {
		return 0
	    } else {
		return 1
	    }
	} else {
	    return 0
	}
    } else {
	return 1
    }
}

if {![info exists alpha::defaultEncoding]} {
    set alpha::defaultEncoding macRoman
}

# This will be over-ridden later.
proc alpha::encodingFor {filename} {
    if {[alpha::inAlphaHierarchy $filename]} {
	return $::alpha::defaultEncoding
    } else {
	return ""
    }
}

# Convert the .tcl files from their default encoding when we source
# them.  Other files (prefs etc.) will be stored in their correct
# encoding, so we only apply this to files in the Alpha:Tcl directory.
if {[lsearch -exact [encoding names] $alpha::defaultEncoding] == -1} {
    catch {alertnote "We don't have a $alpha::defaultEncoding encoding;\
      this will cause serious problems!"}
} else {
    rename source __enc_source
    # This will be replaced later.
    proc source {args} {
	set filename [lindex $args 0]
	if {[llength $args] > 1 \
	  && [lsearch -exact {-rsrc -rsrcid} $filename] != -1} {
	    return [uplevel 1 __enc_source $args]
	}
	if {![file exists $filename] || ![file readable $filename]} {
	    return [uplevel 1 [list __enc_source $filename]]
	}
	set enc [alpha::encodingFor $filename]
	if {![string length $enc]} {
	    uplevel 1 [list __enc_source $filename]
	} else {
	    if {[catch {
		set fileid [open $filename "r"]
		fconfigure $fileid -encoding $enc 
		set contents [read $fileid]
		close $fileid
	    } err]} {
		# This is pretty desperate if we get here!
		alertnote "Error while pre-sourcing $filename : $err"
	    }
	    if {[info tclversion] > 8.3} {
		set oldscript [info script $filename]
	    }
	    uplevel 1 $contents
	    if {[info tclversion] > 8.3} {
		info script $oldscript
	    }
	}
    }
}

if {![llength [info procs alpha::Init]]} {
    proc alpha::Init {} {
	rename alpha::Init {}
	if {[info tclversion] >= 8.0} {
	    # Alpha 8/X/tk all use standard Tcl indices.
	    # 
	    # The AlphaTcl library has its own implementation of unknown,
	    # auto_load, auto_mkindex etc.  This flag turns that off.
	    # AlphaTcl still uses a different version of
	    # auto_mkindex/auto_reset to the core (they contain more
	    # error-checking, for example), but the tclIndex file formats
	    # are identical, and the core's unknown is therefore used.
	    set ::useStandardTclIndices 1
	    if {$::alpha::platform == "alpha"} {
		# This is only used by Alpha 8/X.  When TIP#62 is
		# finalised, we will use that instead.
		if {[catch {package require Tclx 8.0} err]} {
		    alertnote "\[package require Tclx 8.0\] failed:\
		      $err. Proc tracing won't work!"
		}
	    }
	}
    }
}

if {![llength [info procs alpha::PreGuiStartup]]} {
    proc alpha::PreGuiStartup {} {
	rename alpha::PreGuiStartup {}
	set file [file join $::HOME Tcl SystemCode PreGui initAlphaTcl.tcl]
	checkFileExists $file
	uplevel #0 [list source $file]
    }
}

if {![llength [info procs alpha::Startup]]} {
    proc alpha::Startup {} {
	rename alpha::Startup {}
	set file [file join $::HOME Tcl SystemCode AlphaBits.tcl]
	checkFileExists $file
	uplevel #0 [list source $file]
    }
}

alpha::Init

